// ATtiny85 timing using timer1 interrupt

// Started: Feb 27, 2015  G. D. (Joe) Young <jyoung@islandnet.com>
//
// Revised: Mar 01/15 - code output during on time - codppm
//          Mar 06/15 - correct logic, add letter spacing
//                    - try several OSCCAL, OCR1C values - see notes
//          Mar 08/15 - codppmx - 1 MHz crystal processor osc
//                    - needs lfuse changed from 0x62 to 0xeb
//                    - OCR1C theoretically correct value 243 is best
//                    - add trimming adjustment to int service
//          Mar 12/15 - use byte ovflcnt, change timing test to ==
//          Mar 14/15 - 1.8432 MHz xtal s/b exact wi div by 8192, 225
//          Mar 15/15 - clean, allow id completion
//                    - codppmx4 - change o/p to open-drain
//          Mar 28/15 - codppmx5 - don't use pwm mode in timer1
//                    - add txPin open-drain output on pb1 for tx
//                    - modify ledPin to run only on first cycle
//          Mar 29/15 - add sending O per IARU rules, cleanups

#include <avr/io.h>
#include <avr/interrupt.h>

#define THISFOX 2
//#define TRIMCNT 5

// interrupt service variables
volatile word ovflcnt = 0;
word loopovfl;

byte ledPin = 0;
byte txPin = 1;
boolean code = false; // true during code sending
boolean sendm, sendo, msent, sendid, sendsp, cend, ledon;
boolean sendlsp, sendwsp, eltOn;  // loop control of morse output
unsigned long timer;
unsigned long dotLen = 120;  // dot length in millisec (1200/wpm)
byte dash, id;

// change interrupt service to compare match interrupt
ISR( TIM1_COMPA_vect ) {
  ovflcnt += 1;
}

void setupTimer1( ) {
  TCCR1 = bit(CTC1) | 0x0e; // do not use pwm mode
  OCR1C = 224;    // 225 div with 1.8432 MHz xtal
  OCR1A = 224;    // using compare match int
  TIMSK |= bit(OCIE1A);  //add timer1 overflow interrupt
} // setupTimer1 for interrupt generation on overflow

void setup( ) {
  noInterrupts( );    // attempt to sync millis( ) with timer1
  setupTimer1( );
  interrupts( );
  pinMode( ledPin, INPUT );  // setup for open drain o/p
  loopovfl = ovflcnt+1+(THISFOX-1)*12;
  digitalWrite( ledPin, LOW ); // will setup no pullups, 0 in reg.
  digitalWrite( txPin, LOW );
  eltOn = false;
  cend = false;
  ledon = true;
} // setup( )


void loop( ) {
  if( !code && loopovfl == ovflcnt ) {  // start sending
    loopovfl += 12;
    sendm = true;
    dash = 2;
    code = true;
    cend = false;
  }
  if( code && loopovfl == ovflcnt ) {    // end sending
    loopovfl += 48;
    cend = true;    // delay end until id complete
  }
  
  if( code && (sendm || sendo) && !eltOn && ( dash > 0 ) ) {
    timer = millis( ) + 3*dotLen;  // sending a dash
    pinMode( txPin, OUTPUT );
    if( ledon ) pinMode( ledPin, OUTPUT );
    eltOn = true;
    dash--;
    if( dash == 0 ) {
      if( sendm ) msent = true;  // will cause wsp to setup O
      sendm = false;
      sendo = false;
      sendlsp = true;
    } // after m or o, send letter space
  } // if sending m or o

  if( code && eltOn && !sendsp && ( timer < millis( ) ) ) {
    pinMode( txPin, INPUT );
    if( ledon ) pinMode( ledPin, INPUT );
    timer += dotLen;    // timing space following element
    sendsp = true;
    if( sendlsp ) timer += 2*dotLen;
    if( sendwsp ) timer += 6*dotLen;
  } // if element on time is up

  if( code && sendsp && ( timer < millis( ) ) ) {
    eltOn = false;
    sendsp = false;
    if( sendlsp && msent ) {
      msent = false;
      sendo = true;
      dash = 3;
      sendlsp = false;
    } // if m plus space just ended
    if( sendlsp && !msent ) {
      sendid = true;   //cue sending Id
      id = THISFOX;        // this will be read from pins
      sendlsp = false;
    } // if space was letter space and o just sent
    if( sendwsp ) {
      sendm = true;
      dash = 2;
      sendwsp = false;
      if( cend ) {
        code = false;  // turn off sending
        ledon = false;   // and led after first burst
      } // if last word space, end sending
    } // if word space ended, setup to begin again
  } // if sending space 

  if( code && sendid && !eltOn && id > 0 ) {
    timer =  millis( ) + dotLen;
    pinMode( txPin, OUTPUT );
    if( ledon ) pinMode( ledPin, OUTPUT );
    eltOn = true;
    id--;
    if( id == 0 ) {
      sendwsp = true;
      sendid = false;
    }
  } // if sending dots for id
  
  
} // loop( )


