// codppmx8 - ATtiny85 timing using timer1 interrupt

// Started: Feb 27, 2015  G. D. (Joe) Young <jyoung@islandnet.com>
//
// Revised: Mar/Apr/15 - see codppmx7 for earlier version(s)
//          Sep 13/15 - codppmx8 adds callsign identification to all modes. From condev2
//          Sep 16/15 - set callspeed to 40 wpm, same format all modes, detect no callsign
//			Feb 17/16 - correct fault--LED goes steady ON in beacon mode sometimes
//
//
// To permit re-programming the callsign without otherwise modifying the controller
// code, the fox ID (MOE..MO5) and the call are stored in processor's eeprom at 
// location 0, in the form of morse codewords (see condev2).

// An example eeprom entry:
// >>> dump eeprom 
// 0000  07 0f 08 00 0c 02 00 11  02 38 18 12 0d 32 38 ff  |.... ....8.. 28.|
// which corresponds to the text: MOS de VE7BFK/7

// The codeword storage format is used to avoid storing the table of ASCII-to-morse
// An arduino utility, text2code, can be used to produce an intel-hex file for loading
// into eeprom.



// There are 4 modes of operation:
//  fmode = 0  Sprint slow
//          1  Sprint fast
//          2  continuous MO
//          3  normal 1min/4min timing



#include <avr/io.h>
#include <avr/interrupt.h>
#include <EEPROM.h>

#define CALL_INTERVAL 1800  // time interval between sending callsign
//#define THISFOX 1   // now lookup fox number from eeprom

// interrupt service variables
volatile word ovflcnt = 0;
word loopovfl;

const byte ledPin = 0;
const byte txPin = 1;
const byte ainPin = A1;

byte fmode, thisfox;
word onTime, offTime, ledtime, calltime;
boolean code = false;              // true during code sending
boolean sendm, sendo, msent, sendid, sendsp, cend, ledon;
boolean sendlsp, sendwsp, eltOn;   // loop control of morse output
unsigned long timer;
unsigned long dotLen = 221;        // dot length in millisec (1200/wpm)1.8432
unsigned long cdotLen = 55;  // dot length while sending call sign
unsigned long savedotLen;

// change interrupt service to compare match interrupt
ISR( TIM1_COMPA_vect ) {
  ovflcnt += 1;
}

void setupTimer1( ) {
  TCCR1 = bit(CTC1) | 0x0e; // do not use pwm mode
  OCR1C = 224;    // 225 div with 1.8432 MHz xtal
  OCR1A = 224;    // using compare match int
  TIMSK |= bit(OCIE1A);  //add timer1 overflow interrupt
} // setupTimer1 for interrupt generation on overflow

// new functions and variables for callsign insertion
void setupNext( byte );       // functions to setup pointers to code bytes
void setupCall( );
byte setGetBit( );    // code byte parsing, now returns fence bit position 7..0
bool getBit( );
byte codeChar, nextelt, nextBit, bitCnt, nextChar, thisLength;
byte callend;
bool endCode, callsend;



void setup( ) {

  int temp = analogRead( ainPin );      // get operating mode
  fmode = temp>>8;

  onTime = 12;
  offTime = 48;
  if( fmode == 1 ) dotLen -= 64;        // fast sprint
  savedotLen = dotLen;
  if( fmode == 3 ) {
    onTime = 60;
    offTime = 240;
  } // if normal timing
  noInterrupts( );    // attempt to sync millis( ) with timer1
  setupTimer1( );
  interrupts( );
  pinMode( ledPin, INPUT );  // setup for open drain o/p
  digitalWrite( ledPin, LOW ); // will setup no pullups, 0 in reg.
  digitalWrite( txPin, LOW );
  eltOn = false;
  cend = false;
  ledon = true;

  codeChar = EEPROM.read( 2 );  //determine fox number from EE code character
  thisfox = setGetBit( );
  callend = 4;                  //locate end of callsign
  while( EEPROM.read( callend ) != 0xff ) callend++;

  thisLength = 3;
  if( fmode == 2 ) {
    thisfox = 0;              // no fox id for continuous
    thisLength = 2;
    ledtime = ovflcnt + 30;   // 30 sec for led 
  }

// Setup timers
  if( fmode == 0 || fmode == 1 ) {
    loopovfl = ovflcnt+1+(thisfox-1)*onTime;  //indexed start for sprint
  } else {
    loopovfl = ovflcnt + 1;    // no indexing for normal
  } // if sprint modes

  calltime = ovflcnt+2;

} // setup( )


void loop( ) {

  if( !code && loopovfl == ovflcnt ) {  // start sending
    loopovfl += onTime;
//    if( fmode == 3 ) {
//      setupCall( );               // mode 3 send call every start
//    } else {
      dotLen = savedotLen;
      nextChar = 0;
      setupNext( nextChar );      // get first letter of fox code
//    }
    code = true;
    cend = false;
  }

  if( code && loopovfl == ovflcnt && fmode != 2 ) {    // end sending
    loopovfl += offTime;
    cend = true;    // delay end until id complete
  }

  if( fmode==2 && ledtime == ovflcnt ) ledon = false;
  
//  if( fmode != 3 && (ovflcnt-calltime) > CALL_INTERVAL ) {
  if( (ovflcnt-calltime) > CALL_INTERVAL ) {
    calltime = ovflcnt;
    callsend = true;
  }
  
  if( code && !eltOn ) {
    timer = millis( ) + dotLen;            // setup for dot length
    if( getBit( ) ) timer += 2*dotLen;      // sending a dash
    pinMode( txPin, OUTPUT );
    if( ledon ) pinMode( ledPin, OUTPUT );
    eltOn = true;
  } // if not sending, start first element

  if( code && eltOn && !sendsp && ( timer < millis( ) ) ) {
    pinMode( txPin, INPUT );
//    if( ledon ) pinMode( ledPin, INPUT );
    pinMode( ledPin, INPUT );				//Feb 17/16
    timer += dotLen;    // timing space following element
    sendsp = true;
    if( endCode ) {
      endCode = false;
      sendlsp = true;
      timer += 2*dotLen;
      nextChar++;
      if( (nextChar == thisLength) || (nextChar == callend) ) {
        dotLen = savedotLen;
        sendwsp = true;
        timer += 3*dotLen;
        nextChar = 0;          // reset pointer to characters
      } // if last character of this group was just sent
      setupNext( nextChar );
      if( codeChar == 0 ) {
        sendwsp = true;
        timer += 3*dotLen;
        nextChar++;
        setupNext( nextChar );    // if space charcter to be sent
      }
    } // extend space to letter space or word space
  } // if element on time is up

  if( code && sendsp && ( timer < millis( ) ) ) {
    eltOn = false;
    sendsp = false;
    if( sendlsp ) {
      sendlsp = false;
    } // if letter space just ended
    if( sendwsp ) {
      sendwsp = false;
      if( cend ) {
        code = false;  // turn off sending
        ledon = false;   // and led after first burst
      } // if last word space, end sending
//      if( fmode == 2 && callsend ) {
//      if( fmode != 3 && callsend ) {
      if( callsend ) {
        callsend = false;
        setupCall( );
      } // if time to send a callsign
    } // if word space ended, setup to begin again
  } // if sending space 
  
} // loop( )


// support function implementations

void setupNext( byte nextChar ) {
  codeChar = EEPROM.read( nextChar );
  setGetBit( );
} // setup to first code byte (usually for M)

void setupCall( ) {
  nextChar =  4;    // start of callsign in EE
  codeChar = EEPROM.read( nextChar );
  dotLen = cdotLen;
  if( codeChar == 0xff ) {              // allow for suppressing call
    nextChar = 0;
    codeChar = EEPROM.read( nextChar );
    dotLen = savedotLen;
  }
  setGetBit( );
} // setup to first code byte of callsign ID

byte setGetBit( ) {
  bitCnt = 7;
  nextBit = 0b10000000;
  while( (bitCnt > 0) && ((nextBit & codeChar) == 0) ) {
    nextBit = nextBit>>1;
    bitCnt--;
  } // while leading zeros before fence bit
  return bitCnt;
} // setup the code word parsing

bool getBit( ) {
  nextBit = nextBit>>1;
  bitCnt--;
  if( bitCnt == 0 ) {
    endCode = true;
  } // if last bit has just been selected
  return( (nextBit & codeChar) == nextBit );
} // get next code element to send


